import * as React from 'react';

// @flow
// const commonHighlight = 'common-highlight'
function defaultSanitize(string) {
    return string;
}
function escapeRegExpFn(string) {
    // eslint-disable-next-line no-useless-escape
    return string.replace(/[\-\[\]\*\&\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&');
}
/**
 * 合并value
 * @param prev
 * @param next
 * @returns
 */
function combineHighlightWords(prev, next) {
    return [...new Set([].concat(prev.value || []).concat(next.value || []))];
}
/**
 * 合并class
 * @param rest
 * @returns
 */
function combineHighlightClassName(...rest) {
    return [...new Set(rest)].filter(str => str).join(' ');
}
/**
 * 合并两个块
 * @param prevChunk
 * @param nextChunk
 * @param commonClass
 * @returns
 */
function combineTwoChunks(prevChunk, nextChunk) {
    const results = [];
    // 顺序错乱的情况忽略
    if (nextChunk.end <= prevChunk.start) {
        results.push({ ...prevChunk });
        return results;
    }
    /**
     * 如果相邻的两段存在交叉内容这合并，
     * 当只有一种高亮时，直接合并即可
     * 当存在两种类型时，在有交叉的情况下：
     * 1、如果是同类型，那么直接合并
     * 2、如果是不同的类型，获取的部分，相邻的两段只保留自己独立的部分
     */
    if (nextChunk.start < prevChunk.end) {
        const start = prevChunk.start;
        const end = Math.max(prevChunk.end, nextChunk.end);
        const commonS = Math.max(prevChunk.start, nextChunk.start);
        const commonE = Math.min(prevChunk.end, nextChunk.end);
        if (nextChunk.start === prevChunk.start && nextChunk.end === prevChunk.end) {
            results.push({
                ...nextChunk,
                className: combineHighlightClassName(prevChunk.className, nextChunk.className),
                value: combineHighlightWords(prevChunk, nextChunk)
            });
        }
        else if (prevChunk.start === nextChunk.start) {
            results.push({
                start,
                end: commonE,
                highlight: false,
                className: combineHighlightClassName(prevChunk.className, nextChunk.className),
                value: combineHighlightWords(prevChunk, nextChunk)
            }, {
                ...(prevChunk.end < nextChunk.end ? nextChunk : prevChunk),
                start: commonE
            });
        }
        else if (prevChunk.end === nextChunk.end) {
            results.push({
                ...(prevChunk.start < nextChunk.start ? prevChunk : nextChunk),
                end: commonS
            }, {
                start: commonS,
                end: commonE,
                highlight: false,
                className: combineHighlightClassName(prevChunk.className, nextChunk.className),
                value: combineHighlightWords(prevChunk, nextChunk)
            });
        }
        else {
            results.push({
                ...prevChunk,
                start,
                end: commonS
            }, {
                highlight: false,
                start: commonS,
                end: commonE,
                value: combineHighlightWords(prevChunk, nextChunk),
                className: combineHighlightClassName(prevChunk.className, nextChunk.className)
            }, {
                ...nextChunk,
                highlight: false,
                start: commonE,
                end
            });
        }
    }
    else {
        // 完全顺序切独立的两项
        results.push(prevChunk, nextChunk);
    }
    return results;
}
/**
 * 对所有切割的模块进行同类交叉、异类交叉、异类包含处理
 * @return {start:number, end:number}[]
 */
const combineChunks = ({ chunks
// commonClass = commonHighlight
 }) => {
    chunks = chunks
        .sort((first, second) => first.start - second.start)
        .reduce((processedChunks, nextChunk) => {
        // 第一个模块
        if (processedChunks.length === 0) {
            return [nextChunk];
        }
        // 往前找 start 小于 当前 nextChunk.start的片段
        let currentChunk = processedChunks.pop();
        const tempPrevList = [];
        if (currentChunk) {
            tempPrevList.push(currentChunk);
            currentChunk = undefined;
        }
        // const tempPrevList: Chunk[] = [processedChunks.pop()]
        while (tempPrevList[0].start > nextChunk.start) {
            currentChunk = processedChunks.pop();
            if (currentChunk) {
                tempPrevList.push(currentChunk);
                currentChunk = undefined;
            }
        }
        if (tempPrevList.length > 1) {
            // 如果不止一项，这需要递归调用 combineChunks 再进行排序再合并
            processedChunks.push(...combineChunks({ chunks: tempPrevList.concat(nextChunk) }));
        }
        else {
            processedChunks.push(...combineTwoChunks(tempPrevList[0], nextChunk));
        }
        return processedChunks;
    }, []);
    return chunks;
};
/**
 * 截取高亮文本，并获取文本位置
 * @param param0
 */
const execWordPosition = ({ autoEscape, caseSensitive, sanitize = defaultSanitize, searchWords = [], textToHighlight, className, isRichText }) => {
    textToHighlight = sanitize(textToHighlight);
    if (!Array.isArray(searchWords)) {
        searchWords = [];
    }
    /**
     * 对内容进行去重，过滤无效内容
     */
    return ([...new Set([...searchWords])]
        .filter((searchWord) => searchWord) // 移除空的字符串
        // .sort((a, b) => a.length - b.length)
        .reduce((chunks, searchWord) => {
        searchWord = sanitize(searchWord);
        if (autoEscape) {
            // 默认处理特殊字符
            searchWord = escapeRegExpFn(searchWord);
        }
        if (isRichText) {
            searchWord = `(?![^<>]*>)(${searchWord})(?![^<>]*>)`;
        }
        // debugger
        // 正则匹配获取位置
        const regex = new RegExp(searchWord, caseSensitive ? 'g' : 'gi');
        let match;
        while ((match = regex.exec(textToHighlight))) {
            const start = match.index;
            const end = regex.lastIndex;
            // We do not return zero-length matches
            if (end > start) {
                // chunks.push({ highlight: false, start, end, className })
                chunks.push({
                    highlight: false,
                    start,
                    end,
                    className,
                    value: [match[0]]
                });
            }
            // 查看 http://www.regexguru.com/2008/04/watch-out-for-zero-length-matches/
            if (match.index === regex.lastIndex) {
                regex.lastIndex++;
            }
        }
        return chunks;
    }, []));
};
/**
 *
 * 处理words的中多组数据并合并在一起，然后交个combineChunks, combineChunks会利用其下标重新排序
 * @return {start:number, end:number, className: string}[]
 */
const defaultFindChunks = ({ words, activeClasses, ...rest }) => {
    if (typeof words[0] === 'string') {
        const searchWords = words;
        return execWordPosition({
            ...rest,
            searchWords: searchWords,
            className: activeClasses[0] || 'first'
        });
    }
    return words.reduce((oneChunks, searchWords, index) => {
        return oneChunks.concat(execWordPosition({
            ...rest,
            searchWords: [].concat(searchWords),
            className: activeClasses[index]
        }));
    }, []);
};
/**
 * 将原文本拆成包含需要高亮和不需要高亮的全部模块
 * @param chunksToHighlight 所有处理之后需要高亮的模块
 * @param totalLength number
 * @return {start:number, end:number, highlight:boolean, className: string}[]
 */
const fillInChunks = ({ chunksToHighlight, totalLength }) => {
    const allChunks = [];
    const append = (start, end, highlight, chunk) => {
        if (end - start > 0) {
            allChunks.push({
                start,
                end,
                highlight,
                className: chunk?.className,
                key: start,
                value: chunk?.value || []
            });
        }
    };
    if (chunksToHighlight.length === 0) {
        append(0, totalLength, false);
    }
    else {
        let lastIndex = 0;
        chunksToHighlight.forEach((chunk) => {
            append(lastIndex, chunk.start, false, chunk); // 无需高亮的项
            append(chunk.start, chunk.end, true, chunk);
            lastIndex = chunk.end;
        });
        append(lastIndex, totalLength, false);
    }
    return allChunks;
};
/**
 * @return 返回所有的chunks,包括高亮和不需要高亮全部 { start:number, end:number, highlight:boolean, className: string })
 */
const findAll = ({ autoEscape, caseSensitive = false, findChunks = defaultFindChunks, sanitize, isRichText, textToHighlight, words, activeClasses
// commonClass
 }) => (fillInChunks({
    chunksToHighlight: combineChunks({
        chunks: findChunks({
            autoEscape,
            caseSensitive,
            sanitize,
            isRichText,
            textToHighlight,
            words,
            activeClasses
        })
        // commonClass
    }),
    totalLength: textToHighlight ? textToHighlight.length : 0
}));

const Highlighter = ({ textToHighlight, activeClasses, words, autoEscape = true }) => {
    const textChunks = findAll({
        textToHighlight: textToHighlight,
        activeClasses,
        words,
        autoEscape
    });
    return (React.createElement("span", { className: "high-lighter" }, textChunks.map(chunk => {
        const text = textToHighlight.substring(chunk.start, chunk.end);
        if (chunk.highlight) {
            return React.createElement("span", { key: chunk.key, className: chunk.className }, text);
        }
        return React.createElement("span", { key: chunk.key }, text);
    })));
};

export { Highlighter as default };
